@@ -11,7 +11,7 @@ module GHC.ByteCode.Types
   , FFIInfo(..)
   , RegBitmap(..)
   , NativeCallType(..), NativeCallInfo(..), voidTupleReturnInfo, voidPrimCallInfo
-  , ByteOff(..), WordOff(..)
+  , ByteOff(..), WordOff(..), HalfWord(..)
   , UnlinkedBCO(..), BCOPtr(..), BCONPtr(..)
   , ItblEnv, ItblPtr(..)
   , AddrEnv, AddrPtr(..)
@@ -44,7 +44,6 @@ import qualified GHC.Exts.Heap as Heap
 import GHC.Stack.CCS
 import GHC.Cmm.Expr ( GlobalRegSet, emptyRegSet, regSetToList )
 import GHC.Iface.Syntax
-import Language.Haskell.Syntax.Module.Name (ModuleName)
 
 -- -----------------------------------------------------------------------------
 -- Compiled Byte Code
@@ -80,6 +79,12 @@ newtype ByteOff = ByteOff Int
 newtype WordOff = WordOff Int
     deriving (Enum, Eq, Show, Integral, Num, Ord, Real, Outputable)
 
+-- A type for values that are half the size of a word on the target
+-- platform where the interpreter runs (which may be a different
+-- wordsize than the compiler).
+newtype HalfWord = HalfWord Word
+    deriving (Enum, Eq, Show, Integral, Num, Ord, Real, Outputable)
+
 newtype RegBitmap = RegBitmap { unRegBitmap :: Word32 }
     deriving (Enum, Eq, Show, Integral, Num, Ord, Real, Bits, FiniteBits, Outputable)
 
@@ -189,7 +194,7 @@ instance NFData BCONPtr where
 data CgBreakInfo
    = CgBreakInfo
    { cgb_tyvars :: ![IfaceTvBndr] -- ^ Type variables in scope at the breakpoint
-   , cgb_vars   :: ![Maybe (IfaceIdBndr, Word16)]
+   , cgb_vars   :: ![Maybe (IfaceIdBndr, Word)]
    , cgb_resty  :: !IfaceType
    }
 -- See Note [Syncing breakpoint info] in GHC.Runtime.Eval
@@ -237,7 +242,6 @@ data ModBreaks
         -- ^ Array pointing to cost centre for each breakpoint
    , modBreaks_breakInfo :: IntMap CgBreakInfo
         -- ^ info about each breakpoint from the bytecode generator
-   , modBreaks_module :: RemotePtr ModuleName
    }
 
 seqModBreaks :: ModBreaks -> ()
@@ -247,8 +251,7 @@ seqModBreaks ModBreaks{..} =
   rnf modBreaks_vars `seq`
   rnf modBreaks_decls `seq`
   rnf modBreaks_ccs `seq`
-  rnf (fmap seqCgBreakInfo modBreaks_breakInfo) `seq`
-  rnf modBreaks_module
+  rnf (fmap seqCgBreakInfo modBreaks_breakInfo)
 
 -- | Construct an empty ModBreaks
 emptyModBreaks :: ModBreaks
@@ -260,7 +263,6 @@ emptyModBreaks = ModBreaks
    , modBreaks_decls = array (0,-1) []
    , modBreaks_ccs = array (0,-1) []
    , modBreaks_breakInfo = IntMap.empty
-   , modBreaks_module = toRemotePtr nullPtr
    }
 
 {-
