@@ -1,7 +1,6 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE TypeFamilies #-}
-
-{-# OPTIONS_GHC -Wno-incomplete-record-updates #-}
+{-# LANGUAGE BlockArguments #-}
+{-# LANGUAGE TupleSections #-}
 
 {-
 (c) The University of Glasgow 2006
@@ -13,8 +12,6 @@ Desugaring arrow commands
 
 module GHC.HsToCore.Arrows ( dsProcExpr ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
 import GHC.HsToCore.Match
@@ -22,18 +19,17 @@ import GHC.HsToCore.Utils
 import GHC.HsToCore.Monad
 
 import GHC.Hs
-import GHC.Tc.Utils.Zonk
+import GHC.Hs.Syn.Type
 
 -- NB: The desugarer, which straddles the source and Core worlds, sometimes
 --     needs to see source types (newtypes etc), and sometimes not
 --     So WATCH OUT; check each use of split*Ty functions.
 -- Sigh.  This is a pain.
 
-import {-# SOURCE #-} GHC.HsToCore.Expr ( dsExpr, dsLExpr, dsLExprNoLP, dsLocalBinds,
+import {-# SOURCE #-} GHC.HsToCore.Expr ( dsExpr, dsLExpr, dsLocalBinds,
                                           dsSyntaxExpr )
 
 import GHC.Tc.Utils.TcType
-import GHC.Core.Type( splitPiTy )
 import GHC.Core.Multiplicity
 import GHC.Tc.Types.Evidence
 import GHC.Core
@@ -42,6 +38,7 @@ import GHC.Core.Utils
 import GHC.Core.Make
 import GHC.HsToCore.Binds (dsHsWrapper)
 
+
 import GHC.Types.Id
 import GHC.Core.ConLike
 import GHC.Builtin.Types
@@ -52,7 +49,9 @@ import GHC.Utils.Panic
 import GHC.Types.Var.Set
 import GHC.Types.SrcLoc
 import GHC.Data.List.SetOps( assocMaybe )
+import Data.Foldable (toList)
 import Data.List (mapAccumL)
+import Data.List.NonEmpty (NonEmpty(..), nonEmpty)
 import GHC.Utils.Misc
 import GHC.Types.Unique.DSet
 
@@ -75,25 +74,6 @@ mkCmdEnv tc_meths
              the_choice_id  = assocMaybe prs choiceAName
              the_loop_id    = assocMaybe prs loopAName
 
-           -- used as an argument in, e.g., do_premap
-       ; check_lev_poly 3 the_arr_id
-
-           -- used as an argument in, e.g., dsCmdStmt/BodyStmt
-       ; check_lev_poly 5 the_compose_id
-
-           -- used as an argument in, e.g., dsCmdStmt/BodyStmt
-       ; check_lev_poly 4 the_first_id
-
-           -- the result of the_app_id is used as an argument in, e.g.,
-           -- dsCmd/HsCmdArrApp/HsHigherOrderApp
-       ; check_lev_poly 2 the_app_id
-
-           -- used as an argument in, e.g., HsCmdIf
-       ; check_lev_poly 5 the_choice_id
-
-           -- used as an argument in, e.g., RecStmt
-       ; check_lev_poly 4 the_loop_id
-
        ; return (meth_binds, DsCmdEnv {
                arr_id     = Var (unmaybe the_arr_id arrAName),
                compose_id = Var (unmaybe the_compose_id composeAName),
@@ -105,28 +85,13 @@ mkCmdEnv tc_meths
   where
     mk_bind (std_name, expr)
       = do { rhs <- dsExpr expr
-           ; id <- newSysLocalDs Many (exprType rhs)
+           ; id <- newSysLocalDs ManyTy (exprType rhs)
            -- no check needed; these are functions
            ; return (NonRec id rhs, (std_name, id)) }
 
     unmaybe Nothing name = pprPanic "mkCmdEnv" (text "Not found:" <+> ppr name)
     unmaybe (Just id) _  = id
 
-      -- returns the result type of a pi-type (that is, a forall or a function)
-      -- Note that this result type may be ill-scoped.
-    res_type :: Type -> Type
-    res_type ty = res_ty
-      where
-        (_, res_ty) = splitPiTy ty
-
-    check_lev_poly :: Int -- arity
-                   -> Maybe Id -> DsM ()
-    check_lev_poly _     Nothing = return ()
-    check_lev_poly arity (Just id)
-      = dsNoLevPoly (nTimes arity res_type (idType id))
-          (text "In the result of the function" <+> quotes (ppr id))
-
-
 -- arr :: forall b c. (b -> c) -> a b c
 do_arr :: DsCmdEnv -> Type -> Type -> CoreExpr -> CoreExpr
 do_arr ids b_ty c_ty f = mkApps (arr_id ids) [Type b_ty, Type c_ty, f]
@@ -169,18 +134,18 @@ do_premap ids b_ty c_ty d_ty f g
 -- construct CoreExpr for \ (a :: a_ty, b :: b_ty) -> a
 mkFstExpr :: Type -> Type -> DsM CoreExpr
 mkFstExpr a_ty b_ty = do
-    a_var <- newSysLocalDs Many a_ty
-    b_var <- newSysLocalDs Many b_ty
-    pair_var <- newSysLocalDs Many (mkCorePairTy a_ty b_ty)
+    a_var <- newSysLocalDs ManyTy a_ty
+    b_var <- newSysLocalDs ManyTy b_ty
+    pair_var <- newSysLocalDs ManyTy (mkCorePairTy a_ty b_ty)
     return (Lam pair_var
                (coreCasePair pair_var a_var b_var (Var a_var)))
 
 -- construct CoreExpr for \ (a :: a_ty, b :: b_ty) -> b
 mkSndExpr :: Type -> Type -> DsM CoreExpr
 mkSndExpr a_ty b_ty = do
-    a_var <- newSysLocalDs Many a_ty
-    b_var <- newSysLocalDs Many b_ty
-    pair_var <- newSysLocalDs Many (mkCorePairTy a_ty b_ty)
+    a_var <- newSysLocalDs ManyTy a_ty
+    b_var <- newSysLocalDs ManyTy b_ty
+    pair_var <- newSysLocalDs ManyTy (mkCorePairTy a_ty b_ty)
     return (Lam pair_var
                (coreCasePair pair_var a_var b_var (Var b_var)))
 
@@ -193,9 +158,9 @@ because the list of variables is typically not yet defined.
 --      = case v of v { (x1, .., xn) -> body }
 -- But the matching may be nested if the tuple is very big
 
-coreCaseTuple :: UniqSupply -> Id -> [Id] -> CoreExpr -> CoreExpr
-coreCaseTuple uniqs scrut_var vars body
-  = mkTupleCase uniqs vars body scrut_var (Var scrut_var)
+coreCaseTuple :: Id -> [Id] -> CoreExpr -> DsM CoreExpr
+coreCaseTuple scrut_var vars body
+  = mkBigTupleCase vars body (Var scrut_var)
 
 coreCasePair :: Id -> Id -> Id -> CoreExpr -> CoreExpr
 coreCasePair scrut_var var1 var2 body
@@ -211,10 +176,19 @@ mkCorePairExpr e1 e2 = mkCoreTup [e1, e2]
 mkCoreUnitExpr :: CoreExpr
 mkCoreUnitExpr = mkCoreTup []
 
-{-
-The input is divided into a local environment, which is a flat tuple
-(unless it's too big), and a stack, which is a right-nested pair.
-In general, the input has the form
+{- Note [Environment and stack]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+The input is divided into
+
+* A local environment, which is a flat tuple (unless it's too big)
+  The elements of the local environment can be
+  - of kind Type (for ordinary variables), or
+  - of kind Constraint (for dictionaries bound by patterns)
+
+* A stack, which is a right-nested pair.
+  The elements on the stack are always of kind Type.
+
+So in general, the input has the form
 
         ((x1,...,xn), (s1,...(sk,())...))
 
@@ -257,10 +231,9 @@ matchEnvStack   :: [Id]         -- x1..xn
                 -> CoreExpr     -- e
                 -> DsM CoreExpr
 matchEnvStack env_ids stack_id body = do
-    uniqs <- newUniqueSupply
-    tup_var <- newSysLocalDs Many (mkBigCoreVarTupTy env_ids)
-    let match_env = coreCaseTuple uniqs tup_var env_ids body
-    pair_id <- newSysLocalDs Many (mkCorePairTy (idType tup_var) (idType stack_id))
+    tup_var <- newSysLocalDs ManyTy (mkBigCoreVarTupTy env_ids)
+    match_env <- coreCaseTuple tup_var env_ids body
+    pair_id <- newSysLocalDs ManyTy (mkCorePairTy (idType tup_var) (idType stack_id))
     return (Lam pair_id (coreCasePair pair_id tup_var stack_id match_env))
 
 ----------------------------------------------
@@ -276,9 +249,9 @@ matchEnv :: [Id]        -- x1..xn
          -> CoreExpr    -- e
          -> DsM CoreExpr
 matchEnv env_ids body = do
-    uniqs <- newUniqueSupply
-    tup_id <- newSysLocalDs Many (mkBigCoreVarTupTy env_ids)
-    return (Lam tup_id (coreCaseTuple uniqs tup_id env_ids body))
+    tup_id <- newSysLocalDs ManyTy (mkBigCoreVarTupTy env_ids)
+    tup_case <- coreCaseTuple tup_id env_ids body
+    return (Lam tup_id tup_case)
 
 ----------------------------------------------
 --              matchVarStack
@@ -292,7 +265,7 @@ matchVarStack :: [Id] -> Id -> CoreExpr -> DsM (Id, CoreExpr)
 matchVarStack [] stack_id body = return (stack_id, body)
 matchVarStack (param_id:param_ids) stack_id body = do
     (tail_id, tail_code) <- matchVarStack param_ids stack_id body
-    pair_id <- newSysLocalDs Many (mkCorePairTy (idType param_id) (idType tail_id))
+    pair_id <- newSysLocalDs ManyTy (mkCorePairTy (idType param_id) (idType tail_id))
     return (pair_id, coreCasePair pair_id param_id tail_id tail_code)
 
 mkHsEnvStackExpr :: [Id] -> Id -> LHsExpr GhcTc
@@ -321,7 +294,7 @@ dsProcExpr pat (L _ (HsCmdTop (CmdTopTc _unitTy cmd_ty ids) cmd)) = do
     let env_stk_ty = mkCorePairTy env_ty unitTy
     let env_stk_expr = mkCorePairExpr (mkBigCoreVarTup env_ids) mkCoreUnitExpr
     fail_expr <- mkFailExpr (ArrowMatchCtxt ProcExpr) env_stk_ty
-    var <- selectSimpleMatchVarL Many pat
+    var <- selectSimpleMatchVarL ManyTy pat
     match_code <- matchSimply (Var var) (ArrowMatchCtxt ProcExpr) pat env_stk_expr fail_expr
     let pat_ty = hsLPatType pat
     let proc_code = do_premap meth_ids pat_ty env_stk_ty cmd_ty
@@ -368,9 +341,9 @@ dsCmd ids local_vars stack_ty res_ty
     let
         (a_arg_ty, _res_ty') = tcSplitAppTy arrow_ty
         (_a_ty, arg_ty) = tcSplitAppTy a_arg_ty
-    core_arrow <- dsLExprNoLP arrow
+    core_arrow <- dsLExpr arrow
     core_arg   <- dsLExpr arg
-    stack_id   <- newSysLocalDs Many stack_ty
+    stack_id   <- newSysLocalDs ManyTy stack_ty
     core_make_arg <- matchEnvStack env_ids stack_id core_arg
     return (do_premap ids
               (envStackType env_ids stack_ty)
@@ -396,7 +369,7 @@ dsCmd ids local_vars stack_ty res_ty
 
     core_arrow <- dsLExpr arrow
     core_arg   <- dsLExpr arg
-    stack_id   <- newSysLocalDs Many stack_ty
+    stack_id   <- newSysLocalDs ManyTy stack_ty
     core_make_pair <- matchEnvStack env_ids stack_id
           (mkCorePairExpr core_arrow core_arg)
 
@@ -423,8 +396,8 @@ dsCmd ids local_vars stack_ty res_ty (HsCmdApp _ cmd arg) env_ids = do
         stack_ty' = mkCorePairTy arg_ty stack_ty
     (core_cmd, free_vars, env_ids')
              <- dsfixCmd ids local_vars stack_ty' res_ty cmd
-    stack_id <- newSysLocalDs Many stack_ty
-    arg_id <- newSysLocalDsNoLP Many arg_ty
+    stack_id <- newSysLocalDs ManyTy stack_ty
+    arg_id <- newSysLocalDs ManyTy arg_ty
     -- push the argument expression onto the stack
     let
         stack' = mkCorePairExpr (Var arg_id) (Var stack_id)
@@ -449,7 +422,7 @@ dsCmd ids local_vars stack_ty res_ty
         env_ids
   = dsCmdLam ids local_vars stack_ty res_ty pats body env_ids
 
-dsCmd ids local_vars stack_ty res_ty (HsCmdPar _ cmd) env_ids
+dsCmd ids local_vars stack_ty res_ty (HsCmdPar _ _ cmd _) env_ids
   = dsLCmd ids local_vars stack_ty res_ty cmd env_ids
 
 -- D, xs |- e :: Bool
@@ -469,7 +442,7 @@ dsCmd ids local_vars stack_ty res_ty (HsCmdIf _ mb_fun cond then_cmd else_cmd)
        <- dsfixCmd ids local_vars stack_ty res_ty then_cmd
     (core_else, fvs_else, else_ids)
        <- dsfixCmd ids local_vars stack_ty res_ty else_cmd
-    stack_id   <- newSysLocalDs Many stack_ty
+    stack_id   <- newSysLocalDs ManyTy stack_ty
     either_con <- dsLookupTyCon eitherTyConName
     left_con   <- dsLookupDataCon leftDataConName
     right_con  <- dsLookupDataCon rightDataConName
@@ -502,6 +475,8 @@ dsCmd ids local_vars stack_ty res_ty (HsCmdIf _ mb_fun cond then_cmd else_cmd)
         fvs_cond `unionDVarSet` fvs_then `unionDVarSet` fvs_else)
 
 {-
+Note [Desugaring HsCmdCase]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~
 Case commands are treated in much the same way as if commands
 (see above) except that there are more alternatives.  For example
 
@@ -528,74 +503,87 @@ case bodies, containing the following fields:
    bodies with |||.
 -}
 
-dsCmd ids local_vars stack_ty res_ty
-      (HsCmdCase _ exp (MG { mg_alts = L l matches
-                           , mg_ext = MatchGroupTc arg_tys _
-                           , mg_origin = origin }))
-      env_ids = do
-    stack_id <- newSysLocalDs Many stack_ty
-
-    -- Extract and desugar the leaf commands in the case, building tuple
-    -- expressions that will (after tagging) replace these leaves
-
-    let
-        leaves = concatMap leavesMatch matches
-        make_branch (leaf, bound_vars) = do
-            (core_leaf, _fvs, leaf_ids)
-               <- dsfixCmd ids (bound_vars `unionVarSet` local_vars) stack_ty
-                    res_ty leaf
-            return ([mkHsEnvStackExpr leaf_ids stack_id],
-                    envStackType leaf_ids stack_ty,
-                    core_leaf)
-
-    branches <- mapM make_branch leaves
-    either_con <- dsLookupTyCon eitherTyConName
-    left_con <- dsLookupDataCon leftDataConName
-    right_con <- dsLookupDataCon rightDataConName
-    let
-        left_id  = HsConLikeOut noExtField (RealDataCon left_con)
-        right_id = HsConLikeOut noExtField (RealDataCon right_con)
-        left_expr  ty1 ty2 e = noLocA $ HsApp noComments
-                           (noLocA $ mkHsWrap (mkWpTyApps [ty1, ty2]) left_id ) e
-        right_expr ty1 ty2 e = noLocA $ HsApp noComments
-                           (noLocA $ mkHsWrap (mkWpTyApps [ty1, ty2]) right_id) e
-
-        -- Prefix each tuple with a distinct series of Left's and Right's,
-        -- in a balanced way, keeping track of the types.
-
-        merge_branches :: ([LHsExpr GhcTc], Type, CoreExpr)
-                      -> ([LHsExpr GhcTc], Type, CoreExpr)
-                      -> ([LHsExpr GhcTc], Type, CoreExpr) -- AZ
-        merge_branches (builds1, in_ty1, core_exp1)
-                       (builds2, in_ty2, core_exp2)
-          = (map (left_expr in_ty1 in_ty2) builds1 ++
-                map (right_expr in_ty1 in_ty2) builds2,
-             mkTyConApp either_con [in_ty1, in_ty2],
-             do_choice ids in_ty1 in_ty2 res_ty core_exp1 core_exp2)
-        (leaves', sum_ty, core_choices) = foldb merge_branches branches
-
-        -- Replace the commands in the case with these tagged tuples,
-        -- yielding a HsExpr Id we can feed to dsExpr.
-
-        (_, matches') = mapAccumL (replaceLeavesMatch res_ty) leaves' matches
+dsCmd ids local_vars stack_ty res_ty (HsCmdCase _ exp match) env_ids = do
+    stack_id <- newSysLocalDs ManyTy stack_ty
+    (match', core_choices)
+      <- dsCases ids local_vars stack_id stack_ty res_ty match
+    let MG{ mg_ext = MatchGroupTc _ sum_ty _ } = match'
         in_ty = envStackType env_ids stack_ty
 
-    core_body <- dsExpr (HsCase noExtField exp
-                         (MG { mg_alts = L l matches'
-                             , mg_ext = MatchGroupTc arg_tys sum_ty
-                             , mg_origin = origin }))
-        -- Note that we replace the HsCase result type by sum_ty,
-        -- which is the type of matches'
+    core_body <- dsExpr (HsCase (ArrowMatchCtxt ArrowCaseAlt) exp match')
 
     core_matches <- matchEnvStack env_ids stack_id core_body
     return (do_premap ids in_ty sum_ty res_ty core_matches core_choices,
             exprFreeIdsDSet core_body `uniqDSetIntersectUniqSet` local_vars)
 
+{-
+\cases and \case are desugared analogously to a case command (see above).
+For example
+
+        \cases {p1 q1 -> c1; p2 q2 -> c2; p3 q3 -> c3 }
+
+is translated to
+
+        premap (\ ((xs), (e1, (e2,stk))) -> cases e1 e2 of
+                  p1 q1 -> (Left (Left (xs1), stk))
+                  p2 q2 -> Left ((Right (xs2), stk))
+                  p3 q3 -> Right ((xs3), stk))
+        ((c1 ||| c2) ||| c3)
+
+(cases...of is hypothetical notation that works like case...of but with
+multiple scrutinees)
+
+-}
 dsCmd ids local_vars stack_ty res_ty
-      (HsCmdLamCase _ mg@MG { mg_ext = MatchGroupTc [Scaled arg_mult arg_ty] _ }) env_ids = do
-  arg_id <- newSysLocalDs arg_mult arg_ty
-  let case_cmd  = noLocA $ HsCmdCase noExtField (nlHsVar arg_id) mg
-  dsCmdLam ids local_vars stack_ty res_ty [nlVarPat arg_id] case_cmd env_ids
+      (HsCmdLamCase _ lc_variant match@MG { mg_ext = MatchGroupTc {mg_arg_tys = arg_tys} } )
+      env_ids = do
+    arg_ids <- newSysLocalsDs arg_tys
+
+    let match_ctxt = ArrowLamCaseAlt lc_variant
+        pat_vars = mkVarSet arg_ids
+        local_vars' = pat_vars `unionVarSet` local_vars
+        (pat_tys, stack_ty') = splitTypeAt (length arg_tys) stack_ty
+
+    -- construct and desugar a case expression with multiple scrutinees
+    (core_body, free_vars, env_ids') <- trimInput \env_ids -> do
+      stack_id <- newSysLocalDs ManyTy stack_ty'
+      (match', core_choices)
+        <- dsCases ids local_vars' stack_id stack_ty' res_ty match
+
+      let MG{ mg_ext = MatchGroupTc _ sum_ty _ } = match'
+          in_ty = envStackType env_ids stack_ty'
+          discrims = map nlHsVar arg_ids
+      (discrim_vars, matching_code)
+        <- matchWrapper (ArrowMatchCtxt match_ctxt) (Just discrims) match'
+      core_body <- flip (bind_vars discrim_vars) matching_code <$>
+        traverse dsLExpr discrims
+
+      core_matches <- matchEnvStack env_ids stack_id core_body
+      return (do_premap ids in_ty sum_ty res_ty core_matches core_choices,
+              exprFreeIdsDSet core_body `uniqDSetIntersectUniqSet` local_vars')
+
+    param_ids <- mapM (newSysLocalDs ManyTy) pat_tys
+    stack_id' <- newSysLocalDs ManyTy stack_ty'
+
+    -- the expression is built from the inside out, so the actions
+    -- are presented in reverse order
+
+    let -- build a new environment, plus what's left of the stack
+        core_expr = buildEnvStack env_ids' stack_id'
+        in_ty = envStackType env_ids stack_ty
+        in_ty' = envStackType env_ids' stack_ty'
+
+    -- bind the scrutinees to the parameters
+    let match_code = bind_vars arg_ids (map Var param_ids) core_expr
+
+    -- match the parameters against the top of the old stack
+    (stack_id, param_code) <- matchVarStack param_ids stack_id' match_code
+    -- match the old environment and stack against the input
+    select_code <- matchEnvStack env_ids stack_id param_code
+    return (do_premap ids in_ty in_ty' res_ty select_code core_body,
+            free_vars `uniqDSetMinusUniqSet` pat_vars)
+    where
+      bind_vars vars exprs expr = foldr (uncurry bindNonRec) expr $ zip vars exprs
 
 -- D; ys |-a cmd : stk --> t
 -- ----------------------------------
@@ -603,14 +591,14 @@ dsCmd ids local_vars stack_ty res_ty
 --
 --              ---> premap (\ ((xs),stk) -> let binds in ((ys),stk)) c
 
-dsCmd ids local_vars stack_ty res_ty (HsCmdLet _ lbinds@binds body) env_ids = do
+dsCmd ids local_vars stack_ty res_ty (HsCmdLet _ _ lbinds@binds _ body) env_ids = do
     let
         defined_vars = mkVarSet (collectLocalBinders CollWithDictBinders binds)
         local_vars' = defined_vars `unionVarSet` local_vars
 
     (core_body, _free_vars, env_ids')
        <- dsfixCmd ids local_vars' stack_ty res_ty body
-    stack_id <- newSysLocalDs Many stack_ty
+    stack_id <- newSysLocalDs ManyTy stack_ty
     -- build a new environment, plus the stack, using the let bindings
     core_binds <- dsLocalBinds lbinds (buildEnvStack env_ids' stack_id)
     -- match the old environment and stack against the input
@@ -629,12 +617,7 @@ dsCmd ids local_vars stack_ty res_ty (HsCmdLet _ lbinds@binds body) env_ids = do
 --
 --              ---> premap (\ (env,stk) -> env) c
 
-dsCmd ids local_vars stack_ty res_ty do_block@(HsCmdDo stmts_ty
-                                               (L loc stmts))
-                                                                   env_ids = do
-    putSrcSpanDsA loc $
-      dsNoLevPoly stmts_ty
-        (text "In the do-command:" <+> ppr do_block)
+dsCmd ids local_vars stack_ty res_ty (HsCmdDo _ (L _ stmts)) env_ids = do
     (core_stmts, env_ids') <- dsCmdDo ids local_vars res_ty stmts env_ids
     let env_ty = mkBigCoreVarTupTy env_ids
     core_fst <- mkFstExpr env_ty stack_ty
@@ -660,10 +643,10 @@ dsCmd _ local_vars _stack_ty _res_ty (HsCmdArrForm _ op _ _ args) env_ids = do
 
 dsCmd ids local_vars stack_ty res_ty (XCmd (HsWrap wrap cmd)) env_ids = do
     (core_cmd, env_ids') <- dsCmd ids local_vars stack_ty res_ty cmd env_ids
-    core_wrap <- dsHsWrapper wrap
-    return (core_wrap core_cmd, env_ids')
+    dsHsWrapper wrap $ \core_wrap ->
+      return (core_wrap core_cmd, env_ids')
 
-dsCmd _ _ _ _ c _ = pprPanic "dsCmd" (ppr c)
+dsCmd _ _ _ _ c@(HsCmdLam {}) _ = pprPanic "dsCmd" (ppr c)
 
 -- D; ys |-a c : stk --> t      (ys <= xs)
 -- ---------------------
@@ -681,7 +664,7 @@ dsTrimCmdArg local_vars env_ids
     (meth_binds, meth_ids) <- mkCmdEnv ids
     (core_cmd, free_vars, env_ids')
        <- dsfixCmd meth_ids local_vars stack_ty cmd_ty cmd
-    stack_id <- newSysLocalDs Many stack_ty
+    stack_id <- newSysLocalDs ManyTy stack_ty
     trim_code
       <- matchEnvStack env_ids stack_id (buildEnvStack env_ids' stack_id)
     let
@@ -704,9 +687,7 @@ dsfixCmd
                 DIdSet,         -- subset of local vars that occur free
                 [Id])           -- the same local vars as a list, fed back
 dsfixCmd ids local_vars stk_ty cmd_ty cmd
-  = do { putSrcSpanDs (getLocA cmd) $ dsNoLevPoly cmd_ty
-           (text "When desugaring the command:" <+> ppr cmd)
-       ; trimInput (dsLCmd ids local_vars stk_ty cmd_ty cmd) }
+  = trimInput (dsLCmd ids local_vars stk_ty cmd_ty cmd)
 
 -- Feed back the list of local variables actually used a command,
 -- for use as the input tuple of the generated arrow.
@@ -723,7 +704,7 @@ trimInput build_arrow
         (core_cmd, free_vars) <- build_arrow env_ids
         return (core_cmd, free_vars, dVarSetElems free_vars))
 
--- Desugaring for both HsCmdLam and HsCmdLamCase.
+-- Desugaring for both HsCmdLam
 --
 -- D; ys |-a cmd : stk t'
 -- -----------------------------------------------
@@ -747,8 +728,8 @@ dsCmdLam ids local_vars stack_ty res_ty pats body env_ids = do
         (pat_tys, stack_ty') = splitTypeAt (length pats) stack_ty
     (core_body, free_vars, env_ids')
        <- dsfixCmd ids local_vars' stack_ty' res_ty body
-    param_ids <- mapM (newSysLocalDsNoLP Many) pat_tys
-    stack_id' <- newSysLocalDs Many stack_ty'
+    param_ids <- mapM (newSysLocalDs ManyTy) pat_tys
+    stack_id' <- newSysLocalDs ManyTy stack_ty'
 
     -- the expression is built from the inside out, so the actions
     -- are presented in reverse order
@@ -769,6 +750,82 @@ dsCmdLam ids local_vars stack_ty res_ty pats body env_ids = do
     return (do_premap ids in_ty in_ty' res_ty select_code core_body,
             free_vars `uniqDSetMinusUniqSet` pat_vars)
 
+-- Used for case and \case(s)
+-- See Note [Desugaring HsCmdCase]
+dsCases :: DsCmdEnv                               -- arrow combinators
+        -> IdSet                                  -- set of local vars available to this command
+        -> Id                                     -- stack id
+        -> Type                                   -- type of the stack (right-nested tuple)
+        -> Type                                   -- return type of the command
+        -> MatchGroup GhcTc (LHsCmd GhcTc)        -- match group to desugar
+        -> DsM (MatchGroup GhcTc (LHsExpr GhcTc), -- match group with choice tree
+                CoreExpr)                         -- desugared choices
+dsCases ids local_vars stack_id stack_ty res_ty
+        (MG { mg_alts = L l matches
+            , mg_ext = MatchGroupTc arg_tys _ origin
+            }) = do
+
+  -- Extract and desugar the leaf commands in the case, building tuple
+  -- expressions that will (after tagging) replace these leaves
+
+  let leaves = concatMap leavesMatch matches
+      make_branch (leaf, bound_vars) = do
+          (core_leaf, _fvs, leaf_ids)
+             <- dsfixCmd ids (bound_vars `unionVarSet` local_vars) stack_ty
+                  res_ty leaf
+          return ([mkHsEnvStackExpr leaf_ids stack_id],
+                  envStackType leaf_ids stack_ty,
+                  core_leaf)
+
+  branches <- mapM make_branch leaves
+  either_con <- dsLookupTyCon eitherTyConName
+  left_con <- dsLookupDataCon leftDataConName
+  right_con <- dsLookupDataCon rightDataConName
+  void_ty <- mkTyConTy <$> dsLookupTyCon voidTyConName
+  let
+      left_id  = mkConLikeTc (RealDataCon left_con)
+      right_id = mkConLikeTc (RealDataCon right_con)
+      left_expr  ty1 ty2 e = noLocA $ HsApp noComments
+                         (noLocA $ mkHsWrap (mkWpTyApps [ty1, ty2]) left_id ) e
+      right_expr ty1 ty2 e = noLocA $ HsApp noComments
+                         (noLocA $ mkHsWrap (mkWpTyApps [ty1, ty2]) right_id) e
+
+      -- Prefix each tuple with a distinct series of Left's and Right's,
+      -- in a balanced way, keeping track of the types.
+
+      merge_branches :: ([LHsExpr GhcTc], Type, CoreExpr)
+                    -> ([LHsExpr GhcTc], Type, CoreExpr)
+                    -> ([LHsExpr GhcTc], Type, CoreExpr) -- AZ
+      merge_branches (builds1, in_ty1, core_exp1)
+                     (builds2, in_ty2, core_exp2)
+        = (map (left_expr in_ty1 in_ty2) builds1 ++
+              map (right_expr in_ty1 in_ty2) builds2,
+           mkTyConApp either_con [in_ty1, in_ty2],
+           do_choice ids in_ty1 in_ty2 res_ty core_exp1 core_exp2)
+  (leaves', sum_ty, core_choices) <- case nonEmpty branches of
+    Just bs -> return $ foldb merge_branches bs
+    -- when the case command has no alternatives, the sum type from
+    -- Note [Desugaring HsCmdCase] becomes the empty sum type,
+    -- i.e. Void. The choices then effectively become `arr absurd`,
+    -- implemented as `arr \case {}`.
+    Nothing -> ([], void_ty,) . do_arr ids void_ty res_ty <$>
+      dsExpr (HsLamCase EpAnnNotUsed LamCase
+        (MG { mg_alts = noLocA []
+            , mg_ext = MatchGroupTc [Scaled ManyTy void_ty] res_ty (Generated SkipPmc)
+            }))
+
+      -- Replace the commands in the case with these tagged tuples,
+      -- yielding a HsExpr Id we can feed to dsExpr.
+
+  let (_, matches') = mapAccumL (replaceLeavesMatch res_ty) leaves' matches
+
+  -- Note that we replace the MatchGroup result type by sum_ty,
+  -- which is the type of matches'
+  return (MG { mg_alts = L l matches'
+             , mg_ext = MatchGroupTc arg_tys sum_ty origin
+             },
+          core_choices)
+
 {-
 Translation of command judgements of the form
 
@@ -793,12 +850,10 @@ dsCmdDo _ _ _ [] _ = panic "dsCmdDo"
 --
 --              ---> premap (\ (xs) -> ((xs), ())) c
 
-dsCmdDo ids local_vars res_ty [L loc (LastStmt _ body _ _)] env_ids = do
-    putSrcSpanDsA loc $ dsNoLevPoly res_ty
-                         (text "In the command:" <+> ppr body)
+dsCmdDo ids local_vars res_ty [L _ (LastStmt _ body _ _)] env_ids = do
     (core_body, env_ids') <- dsLCmd ids local_vars unitTy res_ty body env_ids
     let env_ty = mkBigCoreVarTupTy env_ids
-    env_var <- newSysLocalDs Many env_ty
+    env_var <- newSysLocalDs ManyTy env_ty
     let core_map = Lam env_var (mkCorePairExpr (Var env_var) mkCoreUnitExpr)
     return (do_premap ids
                         env_ty
@@ -863,7 +918,6 @@ dsCmdStmt ids local_vars out_ids (BodyStmt c_ty cmd _ _) env_ids = do
         out_ty = mkBigCoreVarTupTy out_ids
         before_c_ty = mkCorePairTy in_ty1 out_ty
         after_c_ty = mkCorePairTy c_ty out_ty
-    dsNoLevPoly c_ty empty -- I (Richard E, Dec '16) have no idea what to say here
     snd_fn <- mkSndExpr c_ty out_ty
     return (do_premap ids in_ty before_c_ty out_ty core_mux $
                 do_compose ids before_c_ty after_c_ty out_ty
@@ -901,18 +955,17 @@ dsCmdStmt ids local_vars out_ids (BindStmt _ pat cmd) env_ids = do
     -- projection function
     --          \ (p, (xs2)) -> (zs)
 
-    env_id <- newSysLocalDs Many env_ty2
-    uniqs <- newUniqueSupply
+    env_id <- newSysLocalDs ManyTy env_ty2
     let
        after_c_ty = mkCorePairTy pat_ty env_ty2
        out_ty = mkBigCoreVarTupTy out_ids
-       body_expr = coreCaseTuple uniqs env_id env_ids2 (mkBigCoreVarTup out_ids)
+    body_expr <- coreCaseTuple env_id env_ids2 (mkBigCoreVarTup out_ids)
 
-    fail_expr <- mkFailExpr (StmtCtxt (DoExpr Nothing)) out_ty
-    pat_id    <- selectSimpleMatchVarL Many pat
+    fail_expr <- mkFailExpr (StmtCtxt (HsDoStmt (DoExpr Nothing))) out_ty
+    pat_id    <- selectSimpleMatchVarL ManyTy pat
     match_code
-      <- matchSimply (Var pat_id) (StmtCtxt (DoExpr Nothing)) pat body_expr fail_expr
-    pair_id   <- newSysLocalDs Many after_c_ty
+      <- matchSimply (Var pat_id) (StmtCtxt (HsDoStmt (DoExpr Nothing))) pat body_expr fail_expr
+    pair_id   <- newSysLocalDs ManyTy after_c_ty
     let
         proj_expr = Lam pair_id (coreCasePair pair_id pat_id env_id match_code)
 
@@ -974,12 +1027,11 @@ dsCmdStmt ids local_vars out_ids
 
     -- post_loop_fn = \((later_ids),(env2_ids)) -> (out_ids)
 
-    uniqs <- newUniqueSupply
-    env2_id <- newSysLocalDs Many env2_ty
+    env2_id <- newSysLocalDs ManyTy env2_ty
     let
         later_ty = mkBigCoreVarTupTy later_ids
         post_pair_ty = mkCorePairTy later_ty env2_ty
-        post_loop_body = coreCaseTuple uniqs env2_id env2_ids (mkBigCoreVarTup out_ids)
+    post_loop_body <- coreCaseTuple env2_id env2_ids (mkBigCoreVarTup out_ids)
 
     post_loop_fn <- matchEnvStack later_ids env2_id post_loop_body
 
@@ -1062,7 +1114,7 @@ dsRecCmd ids local_vars stmts later_ids later_rets rec_ids rec_rets = do
 
     -- squash_pair_fn = \ ((env1_ids), ~(rec_ids)) -> (env_ids)
 
-    rec_id <- newSysLocalDs Many rec_ty
+    rec_id <- newSysLocalDs ManyTy rec_ty
     let
         env1_id_set = fv_stmts `uniqDSetMinusUniqSet` rec_id_set
         env1_ids = dVarSetElems env1_id_set
@@ -1072,7 +1124,7 @@ dsRecCmd ids local_vars stmts later_ids later_rets rec_ids rec_rets = do
           where
             selectVar v
                 | v `elemVarSet` rec_id_set
-                  = mkTupleSelector rec_ids v rec_id (Var rec_id)
+                  = mkBigTupleSelector rec_ids v rec_id (Var rec_id)
                 | otherwise = Var v
 
     squash_pair_fn <- matchEnvStack env1_ids rec_id core_body
@@ -1106,7 +1158,7 @@ dsfixCmdStmts
 
 dsfixCmdStmts ids local_vars out_ids stmts
   = trimInput (dsCmdStmts ids local_vars out_ids stmts)
-   -- TODO: Add levity polymorphism check for the resulting expression.
+   -- TODO: Add representation polymorphism check for the resulting expression.
    -- But I (Richard E.) don't know enough about arrows to do so.
 
 dsCmdStmts
@@ -1139,7 +1191,7 @@ dsCmdStmts _ _ _ [] _ = panic "dsCmdStmts []"
 -- Match a list of expressions against a list of patterns, left-to-right.
 
 matchSimplys :: [CoreExpr]              -- Scrutinees
-             -> HsMatchContext GhcRn    -- Match kind
+             -> HsMatchContext GhcTc    -- Match kind
              -> [LPat GhcTc]            -- Patterns they should match
              -> CoreExpr                -- Return this if they all match
              -> CoreExpr                -- Return this if they don't
@@ -1197,11 +1249,12 @@ replaceLeavesGRHS [] _ = panic "replaceLeavesGRHS []"
 
 -- Balanced fold of a non-empty list.
 
-foldb :: (a -> a -> a) -> [a] -> a
-foldb _ [] = error "foldb of empty list"
-foldb _ [x] = x
+foldb :: (a -> a -> a) -> NonEmpty a -> a
+foldb _ (x:|[]) = x
 foldb f xs = foldb f (fold_pairs xs)
   where
-    fold_pairs [] = []
-    fold_pairs [x] = [x]
-    fold_pairs (x1:x2:xs) = f x1 x2:fold_pairs xs
+    fold_pairs (x1:|x2:xs) = f x1 x2 :| keep_empty fold_pairs xs
+    fold_pairs xs          = xs
+
+    keep_empty :: (NonEmpty a -> NonEmpty a) -> [a] -> [a]
+    keep_empty f = maybe [] (toList . f) . nonEmpty
