@@ -4,30 +4,6 @@
  * Copyright (c) The GHC Team, 1994-2002.
  * ---------------------------------------------------------------------------*/
 
-/*
-Note [CBV Functions and the interpreter]
-~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-When the byte code interpreter loads a reference to a value it often
-ends up as a non-tagged pointers *especially* if we already know a value
-is a certain constructor and therefore don't perform an eval on the reference.
-This causes friction with CBV functions which assume
-their value arguments are properly tagged by the caller.
-
-In order to ensure CBV functions still get passed tagged functions we have
-three options:
-a)  Special case the interpreter behaviour into the tag inference analysis.
-    If we assume the interpreter can't properly tag value references the STG passes
-    would then wrap such calls in appropriate evals which are executed at runtime.
-    This would ensure tags by doing additional evals at runtime.
-b)  When the interpreter pushes references for known constructors instead of
-    pushing the objects address add the tag to the value pushed. This is what
-    the NCG backends do.
-c)  When the interpreter pushes a reference inspect the closure of the object
-    and apply the appropriate tag at runtime.
-
-For now we use approach c). Mostly because it's easiest to implement. We also don't
-tag functions as tag inference currently doesn't rely on those being properly tagged.
-*/
 
 #include "rts/PosixSource.h"
 #include "Rts.h"
@@ -158,11 +134,11 @@ tag functions as tag inference currently doesn't rely on those being properly ta
    cap->r.rRet = (retcode);                             \
    return cap;
 
-#define Sp_plusB(n)  ((void *)((StgWord8*)Sp + (ptrdiff_t)(n)))
-#define Sp_minusB(n) ((void *)((StgWord8*)Sp - (ptrdiff_t)(n)))
+#define Sp_plusB(n)  ((void *)(((StgWord8*)Sp) + (n)))
+#define Sp_minusB(n) ((void *)(((StgWord8*)Sp) - (n)))
 
-#define Sp_plusW(n)  (Sp_plusB((ptrdiff_t)(n) * (ptrdiff_t)sizeof(W_)))
-#define Sp_minusW(n) (Sp_minusB((ptrdiff_t)(n) * (ptrdiff_t)sizeof(W_)))
+#define Sp_plusW(n)  (Sp_plusB((n) * sizeof(W_)))
+#define Sp_minusW(n) (Sp_minusB((n) * sizeof(W_)))
 
 #define Sp_addB(n)   (Sp = Sp_plusB(n))
 #define Sp_subB(n)   (Sp = Sp_minusB(n))
@@ -1111,7 +1087,7 @@ run_BCO:
         /* check for a breakpoint on the beginning of a let binding */
         case bci_BRK_FUN:
         {
-            int arg1_brk_array, arg2_array_index, arg3_module_name;
+            int arg1_brk_array, arg2_array_index, arg3_module_uniq;
 #if defined(PROFILING)
             int arg4_cc;
 #endif
@@ -1129,7 +1105,7 @@ run_BCO:
 
             arg1_brk_array      = BCO_GET_LARGE_ARG;
             arg2_array_index    = BCO_NEXT;
-            arg3_module_name    = BCO_GET_LARGE_ARG;
+            arg3_module_uniq    = BCO_GET_LARGE_ARG;
 #if defined(PROFILING)
             arg4_cc             = BCO_GET_LARGE_ARG;
 #else
@@ -1194,7 +1170,7 @@ run_BCO:
                   // continue execution of this BCO when the IO action returns.
                   //
                   // ioAction :: Int#        -- the breakpoint index
-                  //          -> Addr#       -- the module name
+                  //          -> Int#        -- the module uniq
                   //          -> Bool        -- exception?
                   //          -> HValue      -- the AP_STACK, or exception
                   //          -> IO ()
@@ -1208,7 +1184,7 @@ run_BCO:
                   SpW(8)  = (W_)new_aps;
                   SpW(7)  = (W_)False_closure;         // True <=> an exception
                   SpW(6)  = (W_)&stg_ap_ppv_info;
-                  SpW(5)  = (W_)BCO_LIT(arg3_module_name);
+                  SpW(5)  = (W_)BCO_LIT(arg3_module_uniq);
                   SpW(4)  = (W_)&stg_ap_n_info;
                   SpW(3)  = (W_)arg2_array_index;
                   SpW(2)  = (W_)&stg_ap_n_info;
@@ -1250,15 +1226,15 @@ run_BCO:
         }
 
         case bci_PUSH_L: {
-            int o1 = BCO_NEXT;
+            W_ o1 = BCO_GET_LARGE_ARG;
             SpW(-1) = SpW(o1);
             Sp_subW(1);
             goto nextInsn;
         }
 
         case bci_PUSH_LL: {
-            int o1 = BCO_NEXT;
-            int o2 = BCO_NEXT;
+            W_ o1 = BCO_GET_LARGE_ARG;
+            W_ o2 = BCO_GET_LARGE_ARG;
             SpW(-1) = SpW(o1);
             SpW(-2) = SpW(o2);
             Sp_subW(2);
@@ -1266,9 +1242,9 @@ run_BCO:
         }
 
         case bci_PUSH_LLL: {
-            int o1 = BCO_NEXT;
-            int o2 = BCO_NEXT;
-            int o3 = BCO_NEXT;
+            W_ o1 = BCO_GET_LARGE_ARG;
+            W_ o2 = BCO_GET_LARGE_ARG;
+            W_ o3 = BCO_GET_LARGE_ARG;
             SpW(-1) = SpW(o1);
             SpW(-2) = SpW(o2);
             SpW(-3) = SpW(o3);
@@ -1277,42 +1253,42 @@ run_BCO:
         }
 
         case bci_PUSH8: {
-            int off = BCO_NEXT;
+            W_ off = BCO_GET_LARGE_ARG;
             Sp_subB(1);
             *(StgWord8*)Sp = *(StgWord8*)(Sp_plusB(off+1));
             goto nextInsn;
         }
 
         case bci_PUSH16: {
-            int off = BCO_NEXT;
+            W_ off = BCO_GET_LARGE_ARG;
             Sp_subB(2);
             *(StgWord16*)Sp = *(StgWord16*)(Sp_plusB(off+2));
             goto nextInsn;
         }
 
         case bci_PUSH32: {
-            int off = BCO_NEXT;
+            W_ off = BCO_GET_LARGE_ARG;
             Sp_subB(4);
             *(StgWord32*)Sp = *(StgWord32*)(Sp_plusB(off+4));
             goto nextInsn;
         }
 
         case bci_PUSH8_W: {
-            int off = BCO_NEXT;
+            W_ off = BCO_GET_LARGE_ARG;
             *(StgWord*)(Sp_minusW(1)) = *(StgWord8*)(Sp_plusB(off));
             Sp_subW(1);
             goto nextInsn;
         }
 
         case bci_PUSH16_W: {
-            int off = BCO_NEXT;
+            W_ off = BCO_GET_LARGE_ARG;
             *(StgWord*)(Sp_minusW(1)) = *(StgWord16*)(Sp_plusB(off));
             Sp_subW(1);
             goto nextInsn;
         }
 
         case bci_PUSH32_W: {
-            int off = BCO_NEXT;
+            W_ off = BCO_GET_LARGE_ARG;
             *(StgWord*)(Sp_minusW(1)) = *(StgWord32*)(Sp_plusB(off));
             Sp_subW(1);
             goto nextInsn;
@@ -1320,48 +1296,13 @@ run_BCO:
 
         case bci_PUSH_G: {
             W_ o1 = BCO_GET_LARGE_ARG;
-            StgClosure *tagged_obj = (StgClosure*) BCO_PTR(o1);
-
-            tag_push_g:
-            ASSERT(LOOKS_LIKE_CLOSURE_PTR((StgClosure*) tagged_obj));
-            // Here we make sure references we push are tagged.
-            // See Note [CBV Functions and the interpreter] in Info.hs
-
-            //Safe some memory reads if we already have a tag.
-            if(GET_CLOSURE_TAG(tagged_obj) == 0) {
-                StgClosure *obj = UNTAG_CLOSURE(tagged_obj);
-                switch ( get_itbl(obj)->type ) {
-                    case IND:
-                    case IND_STATIC:
-                    {
-                        tagged_obj = ACQUIRE_LOAD(&((StgInd*)obj)->indirectee);
-                        goto tag_push_g;
-                    }
-                    case CONSTR:
-                    case CONSTR_1_0:
-                    case CONSTR_0_1:
-                    case CONSTR_2_0:
-                    case CONSTR_1_1:
-                    case CONSTR_0_2:
-                    case CONSTR_NOCAF:
-                        // The value is already evaluated, so we can just return it. However,
-                        // before we do, we MUST ensure that the pointer is tagged, because we
-                        // might return to a native `case` expression, which assumes the returned
-                        // pointer is tagged so it can use the tag to select an alternative.
-                        tagged_obj = tagConstr(obj);
-                        break;
-                    default:
-                        break;
-                }
-            }
-
-            SpW(-1) = (W_) tagged_obj;
+            SpW(-1) = BCO_PTR(o1);
             Sp_subW(1);
             goto nextInsn;
         }
 
         case bci_PUSH_ALTS_P: {
-            int o_bco  = BCO_GET_LARGE_ARG;
+            W_ o_bco  = BCO_GET_LARGE_ARG;
             Sp_subW(2);
             SpW(1) = BCO_PTR(o_bco);
             SpW(0) = (W_)&stg_ctoi_R1p_info;
@@ -1374,7 +1315,7 @@ run_BCO:
         }
 
         case bci_PUSH_ALTS_N: {
-            int o_bco  = BCO_GET_LARGE_ARG;
+            W_ o_bco  = BCO_GET_LARGE_ARG;
             SpW(-2) = (W_)&stg_ctoi_R1n_info;
             SpW(-1) = BCO_PTR(o_bco);
             Sp_subW(2);
@@ -1387,7 +1328,7 @@ run_BCO:
         }
 
         case bci_PUSH_ALTS_F: {
-            int o_bco  = BCO_GET_LARGE_ARG;
+            W_ o_bco  = BCO_GET_LARGE_ARG;
             SpW(-2) = (W_)&stg_ctoi_F1_info;
             SpW(-1) = BCO_PTR(o_bco);
             Sp_subW(2);
@@ -1400,7 +1341,7 @@ run_BCO:
         }
 
         case bci_PUSH_ALTS_D: {
-            int o_bco  = BCO_GET_LARGE_ARG;
+            W_ o_bco  = BCO_GET_LARGE_ARG;
             SpW(-2) = (W_)&stg_ctoi_D1_info;
             SpW(-1) = BCO_PTR(o_bco);
             Sp_subW(2);
@@ -1413,7 +1354,7 @@ run_BCO:
         }
 
         case bci_PUSH_ALTS_L: {
-            int o_bco  = BCO_GET_LARGE_ARG;
+            W_ o_bco  = BCO_GET_LARGE_ARG;
             SpW(-2) = (W_)&stg_ctoi_L1_info;
             SpW(-1) = BCO_PTR(o_bco);
             Sp_subW(2);
@@ -1426,7 +1367,7 @@ run_BCO:
         }
 
         case bci_PUSH_ALTS_V: {
-            int o_bco  = BCO_GET_LARGE_ARG;
+            W_ o_bco  = BCO_GET_LARGE_ARG;
             SpW(-2) = (W_)&stg_ctoi_V_info;
             SpW(-1) = BCO_PTR(o_bco);
             Sp_subW(2);
@@ -1439,9 +1380,9 @@ run_BCO:
         }
 
         case bci_PUSH_ALTS_T: {
-            int o_bco = BCO_GET_LARGE_ARG;
+            W_ o_bco = BCO_GET_LARGE_ARG;
             W_ tuple_info = (W_)BCO_LIT(BCO_GET_LARGE_ARG);
-            int o_tuple_bco = BCO_GET_LARGE_ARG;
+            W_ o_tuple_bco = BCO_GET_LARGE_ARG;
 
 #if defined(PROFILING)
             SpW(-1) = (W_)cap->r.rCCCS;
@@ -1585,30 +1526,30 @@ run_BCO:
         }
 
         case bci_PUSH_UBX8: {
-            int o_lit = BCO_GET_LARGE_ARG;
+            W_ o_lit = BCO_GET_LARGE_ARG;
             Sp_subB(1);
             *(StgWord8*)Sp = *(StgWord8*)(literals+o_lit);
             goto nextInsn;
         }
 
         case bci_PUSH_UBX16: {
-            int o_lit = BCO_GET_LARGE_ARG;
+            W_ o_lit = BCO_GET_LARGE_ARG;
             Sp_subB(2);
             *(StgWord16*)Sp = *(StgWord16*)(literals+o_lit);
             goto nextInsn;
         }
 
         case bci_PUSH_UBX32: {
-            int o_lit = BCO_GET_LARGE_ARG;
+            W_ o_lit = BCO_GET_LARGE_ARG;
             Sp_subB(4);
             *(StgWord32*)Sp = *(StgWord32*)(literals+o_lit);
             goto nextInsn;
         }
 
         case bci_PUSH_UBX: {
-            int i;
-            int o_lits = BCO_GET_LARGE_ARG;
-            int n_words = BCO_NEXT;
+            W_ i;
+            W_ o_lits = BCO_GET_LARGE_ARG;
+            W_ n_words = BCO_GET_LARGE_ARG;
             Sp_subW(n_words);
             for (i = 0; i < n_words; i++) {
                 SpW(i) = (W_)BCO_LIT(o_lits+i);
@@ -1617,10 +1558,10 @@ run_BCO:
         }
 
         case bci_SLIDE: {
-            int n  = BCO_NEXT;
-            int by = BCO_NEXT;
+            W_ n  = BCO_GET_LARGE_ARG;
+            W_ by = BCO_GET_LARGE_ARG;
             /* a_1, .. a_n, b_1, .. b_by, s => a_1, .. a_n, s */
-            while(--n >= 0) {
+            while(n-- > 0) {
                 SpW(n+by) = SpW(n);
             }
             Sp_addW(by);
@@ -1629,7 +1570,7 @@ run_BCO:
         }
 
         case bci_ALLOC_AP: {
-            int n_payload = BCO_NEXT;
+            StgHalfWord n_payload = BCO_GET_LARGE_ARG;
             StgAP *ap = (StgAP*)allocate(cap, AP_sizeW(n_payload));
             SpW(-1) = (W_)ap;
             ap->n_args = n_payload;
@@ -1642,7 +1583,7 @@ run_BCO:
         }
 
         case bci_ALLOC_AP_NOUPD: {
-            int n_payload = BCO_NEXT;
+            StgHalfWord n_payload = BCO_GET_LARGE_ARG;
             StgAP *ap = (StgAP*)allocate(cap, AP_sizeW(n_payload));
             SpW(-1) = (W_)ap;
             ap->n_args = n_payload;
@@ -1656,8 +1597,8 @@ run_BCO:
 
         case bci_ALLOC_PAP: {
             StgPAP* pap;
-            int arity = BCO_NEXT;
-            int n_payload = BCO_NEXT;
+            StgHalfWord arity = BCO_GET_LARGE_ARG;
+            StgHalfWord n_payload = BCO_GET_LARGE_ARG;
             pap = (StgPAP*)allocate(cap, PAP_sizeW(n_payload));
             SpW(-1) = (W_)pap;
             pap->n_args = n_payload;
@@ -1670,11 +1611,11 @@ run_BCO:
         }
 
         case bci_MKAP: {
-            int i;
-            int stkoff = BCO_NEXT;
-            int n_payload = BCO_NEXT;
+            StgHalfWord i;
+            W_ stkoff = BCO_GET_LARGE_ARG;
+            StgHalfWord n_payload = BCO_GET_LARGE_ARG;
             StgAP* ap = (StgAP*)SpW(stkoff);
-            ASSERT((int)ap->n_args == n_payload);
+            ASSERT(ap->n_args == n_payload);
             ap->fun = (StgClosure*)SpW(0);
 
             // The function should be a BCO, and its bitmap should
@@ -1694,11 +1635,11 @@ run_BCO:
         }
 
         case bci_MKPAP: {
-            int i;
-            int stkoff = BCO_NEXT;
-            int n_payload = BCO_NEXT;
+            StgHalfWord i;
+            W_ stkoff = BCO_GET_LARGE_ARG;
+            StgHalfWord n_payload = BCO_GET_LARGE_ARG;
             StgPAP* pap = (StgPAP*)SpW(stkoff);
-            ASSERT((int)pap->n_args == n_payload);
+            ASSERT(pap->n_args == n_payload);
             pap->fun = (StgClosure*)SpW(0);
 
             // The function should be a BCO
@@ -1722,8 +1663,8 @@ run_BCO:
 
         case bci_UNPACK: {
             /* Unpack N ptr words from t.o.s constructor */
-            int i;
-            int n_words = BCO_NEXT;
+            W_ i;
+            W_ n_words = BCO_GET_LARGE_ARG;
             StgClosure* con = UNTAG_CLOSURE((StgClosure*)SpW(0));
             Sp_subW(n_words);
             for (i = 0; i < n_words; i++) {
@@ -1733,9 +1674,9 @@ run_BCO:
         }
 
         case bci_PACK: {
-            int i;
-            int o_itbl         = BCO_GET_LARGE_ARG;
-            int n_words        = BCO_NEXT;
+            W_ i;
+            W_ o_itbl         = BCO_GET_LARGE_ARG;
+            W_ n_words        = BCO_GET_LARGE_ARG;
             StgInfoTable* itbl = INFO_PTR_TO_STRUCT((StgInfoTable *)BCO_LIT(o_itbl));
             int request        = CONSTR_sizeW( itbl->layout.payload.ptrs,
                                                itbl->layout.payload.nptrs );
@@ -2065,9 +2006,9 @@ run_BCO:
         }
 
         case bci_SWIZZLE: {
-            int stkoff = BCO_NEXT;
-            signed short n = (signed short)(BCO_NEXT);
-            SpW(stkoff) += (W_)n;
+            W_ stkoff = BCO_GET_LARGE_ARG;
+            StgInt n = BCO_GET_LARGE_ARG;
+            (*(StgInt*)(Sp_plusW(stkoff))) += n;
             goto nextInsn;
         }
 
@@ -2079,7 +2020,7 @@ run_BCO:
 
         case bci_CCALL: {
             void *tok;
-            int stk_offset            = BCO_NEXT;
+            W_ stk_offset             = BCO_GET_LARGE_ARG;
             int o_itbl                = BCO_GET_LARGE_ARG;
             int flags                 = BCO_NEXT;
             bool interruptible        = flags & 0x1;
@@ -2115,7 +2056,7 @@ run_BCO:
             uint32_t nargs = cif->nargs;
             uint32_t ret_size;
             uint32_t i;
-            int j;
+            W_ j;
             StgPtr p;
             W_ ret[2];                  // max needed
             W_ *arguments[stk_offset];  // max needed
