@@ -1890,9 +1890,6 @@ static void normaliseRtsOpts (void)
     if (RtsFlags.GcFlags.maxHeapSize != 0 &&
         RtsFlags.GcFlags.heapSizeSuggestion >
         RtsFlags.GcFlags.maxHeapSize) {
-        errorBelch("Maximum heap size (-M) is smaller than suggested heap size (-H)\n"
-                   "Setting maximum heap size to suggested heap size ( %" FMT_Word64 " )",
-                   (StgWord64) RtsFlags.GcFlags.maxHeapSize * (StgWord64) BLOCK_SIZE);
         RtsFlags.GcFlags.maxHeapSize = RtsFlags.GcFlags.heapSizeSuggestion;
     }
 
@@ -2204,13 +2201,14 @@ static void read_debug_flags(const char* arg)
     }
     // -Dx also turns on -v.  Use -l to direct trace
     // events to the .eventlog file instead.
-    RtsFlags.TraceFlags.tracing = TRACE_STDERR;
-
-   // sanity implies zero_on_gc
-   if(RtsFlags.DebugFlags.sanity){
-        RtsFlags.DebugFlags.zero_on_gc = true;
-   }
+    if (RtsFlags.TraceFlags.tracing == TRACE_NONE) {
+        RtsFlags.TraceFlags.tracing = TRACE_STDERR;
+    }
 
+    // sanity implies zero_on_gc
+    if(RtsFlags.DebugFlags.sanity){
+         RtsFlags.DebugFlags.zero_on_gc = true;
+    }
 }
 #endif
 
